//
//  ap.cpp v1.9
//  Copyright 1995 by Matthew Scott
//  All Rights Reserved.
//
//  takes a text file and creates a msg.69 file
//      subject field contains a custom subject
//      comment field contains filename and "part x"
//      address is pst.xxx
//
//  numerous revisions from v1.0 -> v1.9 -- see docs
//

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include "getopt.h"

#define FNLEN		256
#define FIELDLEN	8192

#define LABELSIZE   30
#define SUBJECTSIZE 80
#define SIGSIZE     90
#define LINESIZE    512

#define CTRLA5      "5"

typedef char fn_t[FNLEN + 1];
typedef char label_t[LABELSIZE + 1];
typedef char subject_t[SUBJECTSIZE + 1];
typedef char sig_t[SIGSIZE + 1];
typedef char line_t[LINESIZE + 1];

int msglimit;

struct message_field_t {
	char fieldid;
	char *fieldtext;
	message_field_t *next;
};

struct message_t {
    label_t number;
	message_field_t *first;
};

void eprintf(int errorlevel, const char *l, ...);
void putstr(FILE *f, const char *str);
void put_message(FILE *f, const message_t& msg);
void kill_message(message_t& msg);
void reset_message(message_t& msg, int part);
void add_field(message_t& msg, char fieldid, const char * text);


#define STRLESSER -1
#define STREQUAL 0
#define STRGREATER 1

fn_t    input_filename;
fn_t    output_filename;
FILE*   input_file;
FILE*   output_file;

subject_t   subject;
label_t     room;
label_t     author;
label_t     node;
label_t     region;
label_t     address;
label_t     realname;
label_t     title;
label_t     surname;
label_t     country;
label_t     twitregion;
label_t     twitcountry;
label_t     phone;
sig_t       bbssig;
sig_t       usersig;
label_t     publicto;
label_t     group;

bool yell;
bool ctrla5;

const char* software = "AutoPoster";
const char* version = "1.9";

int main(int argc, char **argv)
{
	// initialize the module variables
    strcpy( subject,    ""                      );
    strcpy( room,       ""                      );
    strcpy( author,     "AutoPoster"            );
    strcpy( node,       "AutoPoster"            );
    strcpy( region,     "Cyberspace"            );
    strcpy( address,    "pst.cyb"               );
    strcpy( realname,   ""                      );
    strcpy( title,      ""                      );
    strcpy( surname,    ""                      );
    strcpy( country,    "USA"                   );
    strcpy( twitregion, ""                      );
    strcpy( twitcountry,""                      );
    strcpy( phone,      "no phone"              );
    strcpy( bbssig,     ""                      );
    strcpy( usersig,    ""                      );
    strcpy( publicto,   ""                      );
    strcpy( group,      ""                      );

	input_file = stdin;
	output_file = stdout;
    yell = false;
    ctrla5 = false;
    msglimit = 8100;

	static option long_opts[] = {
        { "yell",           0,  0,  'y' },
        { "input",          1,  0,  'i' },
        { "output",         1,  0,  'o' },
        { "help",           0,  0,  'h' },  // also -?
        { "msglimit",       1,  0,  'm' },
        { "ctrla5",         0,  0,  '5' },

        { "room",           1,  0,  'R' },
        { "subject",        1,  0,  'B' },
        { "publicto",       1,  0,  't' },
        { "title",          1,  0,  'N' },
        { "author",         1,  0,  'A' },
        { "surname",        1,  0,  'n' },
        { "realname",       1,  0,  '!' },
        { "address",        1,  0,  'P' },
        { "node",           1,  0,  'O' },
        { "region",         1,  0,  'r' },
        { "twitregion",     1,  0,  'J' },
        { "country",        1,  0,  'Q' },
        { "twitcountry",    1,  0,  'j' },
        { "phone",          1,  0,  'H' },
        { "usersig",        1,  0,  '_' },
        { "bbssig",         1,  0,  '.' },
        { "group",          1,  0,  'G' },

        { 0,                0,  0,  0   }
    };
    static char *short_opts = "yi:o:h?m:5R:B:t:N:A:n:!:P:O:r:J:Q:j:H:_:.:G:";

    const char* helptext =
        "autopost [@cfgfile] [options]\n"
        "   --yell              or -y\n"
        "   --input=filename    or -i filename (default: stdin)\n"
        "   --output=filename   or -o filename (default: stdout)\n"
        "   --help              or -? or -h\n"
        "   --msglimit=limit    or -m limit\n"
        "   --ctrla5            or -5\n"
        "\n"
        "   --room=text         or -R text\n"
        "   --subject=text      or -B text\n"
        "   --publicto=text     or -t text\n"
        "   --title=text        or -N text\n"
        "   --author=text       or -A text\n"
        "   --surname=text      or -n text\n"
        "   --realname=text     or -! text\n"
        "   --address=text      or -P text\n"
        "   --node=text         or -O text\n"
        "   --region=text       or -r text\n"
        "   --twitregion=text   or -J text\n"
        "   --country=text      or -Q text\n"
        "   --twitcountry=text  or -j text\n"
        "   --phone=text        or -H text\n"
        "   --usersig=text      or -_ text\n"
        "   --bbssig=text       or -. text\n"
        "   --group=text        or -G text\n";

	int c;

	while (1) {
		int option_index = 0;

		c = getopt_long(argc, argv, short_opts, long_opts, &option_index);
		if (c == EOF)
			break;

		const char *optname;
		switch (c) {
			case 'i':
				strncpy(input_filename, optarg, FNLEN);
				input_file = fopen(input_filename, "r");
				if (input_file == NULL)
					eprintf(1, "autopost: couldn't open input file %s\n", input_filename);
				break;

			case 'o':
				strncpy(output_filename, optarg, FNLEN);
				output_file = fopen(output_filename, "wb");
				if (output_file == NULL)
					eprintf(1, "autopost: couldn't open output file %s\n", output_filename);
				break;

            case 'm':
                msglimit = atoi(optarg);
                if (msglimit < LINESIZE)
                    eprintf(1, "autopost: msglimit must be at least %d\n", LINESIZE);
                break;

			case 'y':
                yell = true;
				break;

            case '5':
                ctrla5 = true;
                break;

            case 'R':
                strncpy(room, optarg, LABELSIZE);
                break;

            case 'B':
                strncpy(subject, optarg, SUBJECTSIZE);
                break;

            case 't':
                strncpy(publicto, optarg, LABELSIZE);
                break;

            case 'N':
                strncpy(title, optarg, LABELSIZE);
                break;

            case 'A':
                strncpy(author, optarg, LABELSIZE);
                break;

            case 'n':
                strncpy(surname, optarg, LABELSIZE);
                break;

            case '!':
                strncpy(realname, optarg, LABELSIZE);
                break;

            case 'P':
                strncpy(address, optarg, LABELSIZE);
                break;

            case 'O':
                strncpy(node, optarg, LABELSIZE);
                break;

            case 'r':
                strncpy(region, optarg, LABELSIZE);
                break;

            case 'J':
                strncpy(twitregion, optarg, LABELSIZE);
                break;

            case 'Q':
                strncpy(country, optarg, LABELSIZE);
                break;

            case 'j':
                strncpy(twitcountry, optarg, LABELSIZE);
                break;

            case 'H':
                strncpy(phone, optarg, LABELSIZE);
                break;

            case '_':
                strncpy(usersig, optarg, SIGSIZE);
                break;

            case '.':
                strncpy(bbssig, optarg, SIGSIZE);
                break;

            case 'G':
                strncpy(group, optarg, LABELSIZE);
                break;

			case 'h':
			case '?':
                eprintf(0, helptext);
				break;

		} // switch
	} // while

    if (yell) {
        fprintf(stderr, "%s v%s: compiled on %s %s\n", software, version,
            __DATE__, __TIME__);
    }

    if (*room == 0)
		eprintf(1, "autopost: you didn't specify what room you want to post to\n");

    if (yell) {
        fprintf(stderr,
            "msglimit:      %d\n"
            "room:          '%s'\n"
            "subject:       '%s'\n"
            "publicto:      '%s'\n"
            "title:         '%s'\n"
            "author:        '%s'\n"
            "surname:       '%s'\n"
            "realname:      '%s'\n"
            "address:       '%s'\n"
            "node:          '%s'\n"
            "region:        '%s'\n"
            "twitregion:    '%s'\n"
            "country:       '%s'\n"
            "twitcountry:   '%s'\n"
            "phone:         '%s'\n"
            "usersig:       '%s'\n"
            "bbssig:        '%s'\n"
            "group:         '%s'\n",
                msglimit,
                room,
                subject,
                publicto,
                title,
                author,
                surname,
                realname,
                address,
                node,
                region,
                twitregion,
                country,
                twitcountry,
                phone,
                usersig,
                bbssig,
                group
        );
    }

	int partnumber = 1;

    char* temptext = new char[FIELDLEN + 1];
    char* templine = new line_t;
	if (!temptext || !templine)
		eprintf(1, "autopost: not enough memory available\n");

    fprintf(output_file, "%s", address);
	fputc(0, output_file);
    putstr(output_file, room);

	message_t msg;
	reset_message(msg, partnumber);
    *temptext = 0;
    if (ctrla5)
        strcat(temptext, CTRLA5);      // ^A5
    while ((fgets(templine, LINESIZE, input_file)) != NULL) {
		if (templine[strlen(templine)-1] == '\n')
			templine[strlen(templine)-1] = 0;

        int lenwithextras = 1 + strlen(templine);
        if (!ctrla5)
            ++lenwithextras;        // extra space at beg. of line.

        if (lenwithextras + strlen(temptext) > msglimit) {
			add_field(msg, 'M', temptext);
			put_message(output_file, msg);
			kill_message(msg);
			*temptext = 0;
            partnumber++;
            reset_message(msg, partnumber);
            if (ctrla5)
                strcat(temptext, CTRLA5);      // ^A5
		}

        if (!ctrla5)
            strcat(temptext, " ");

        strcat(temptext, templine);
        strcat(temptext, "\n");
    } // while

    add_field(msg, 'M', temptext);
    put_message(output_file, msg);
    kill_message(msg);
    fputc(0, output_file);
	
    delete templine;
    delete temptext;

	return 0;
			
} // main




void reset_message(message_t& msg, int part)
{
    if (yell)
        fprintf(stderr, "On part #%d of %s (%s)\n", part, subject, input_filename);

    time_t curtime;
    time(&curtime);
    char miscstring[81];

    msg.first = NULL;

    sprintf(msg.number, "%d", part);
    add_field(msg, 'A', author);
    sprintf(miscstring, "%lu", curtime);
	add_field(msg, 'D', miscstring);
    add_field(msg, 'O', node);
    sprintf(miscstring, "%d", part);
	add_field(msg, 'S', miscstring);
    add_field(msg, 'o', region);
    sprintf(miscstring, "%s", address);
    add_field(msg, 'P', miscstring);
    add_field(msg, 'Q', country);
    sprintf(miscstring, "%s v%s", software, version);
    add_field(msg, 's', miscstring);

    if (*phone)             add_field(msg, 'H', phone);
    if (*subject)           add_field(msg, 'B', subject);
    if (*realname)          add_field(msg, '!', realname);
    if (*title)             add_field(msg, 'N', title);
    if (*surname)           add_field(msg, 'n', surname);
    if (*twitregion)        add_field(msg, 'J', twitregion);
    if (*twitcountry)       add_field(msg, 'j', twitcountry);
    if (*bbssig)            add_field(msg, '.', bbssig);
    if (*usersig)           add_field(msg, '_', usersig);
    if (*publicto)          add_field(msg, 't', publicto);
    if (*group)             add_field(msg, 'G', group);

	if (*input_filename)
		sprintf(miscstring, "%s Part %d", input_filename, part);
	else
		sprintf(miscstring, "(stdin) Part %d", part);
	add_field(msg, '%', miscstring);

} // reset_message






void add_field(message_t& msg, char fieldid, const char * text)
{
    message_field_t *chainfield = msg.first;
    if (chainfield != NULL)
        while (chainfield->next != NULL)
            chainfield = chainfield->next;

    message_field_t *field = new message_field_t;
	if (!field)
        eprintf(1, "autopost: not enough memory in add_field()\n");
    field->fieldtext = new char[strlen(text) + 1];
	if (!field->fieldtext)
        eprintf(1, "autopost: not enough memory in add_field()\n");
    strcpy(field->fieldtext, text);

    field->fieldid = fieldid;
    field->next = NULL;

    if (chainfield != NULL)
        chainfield->next = field;
    else
        msg.first = field;

} // add_field





void eprintf(int errorlevel, const char *l, ...)
{
	char dlin[255];
	va_list ap;
	va_start(ap, l);
	vsprintf(dlin, l, ap);
	va_end(ap);
	
    if (errorlevel)
        fputs(dlin, stderr);
    else
        fputs(dlin, stdout);
	exit(errorlevel);
} //eprintf





void putstr(FILE *f, const char *str)
{
	const char *s2 = str;
	
	do {
		fputc(*s2, f);
	} while (*(s2++));
} // putstr




void put_message(FILE *f, const message_t& msg)
{
	if (msg.first == NULL)
		return;
		
	// put the 0xff string for start of message
	fputc(0xff, f);
	fputc(0, f);
	
	putstr(f, msg.number);
	message_field_t *current = msg.first;
	
	while (current != NULL) {
		fputc(current->fieldid, f);
		putstr(f, current->fieldtext);
		current = current->next;
	} // while
} // put_message



void kill_message(message_t& msg)
{
	while (msg.first != NULL) {
        message_field_t *temp = msg.first->next;
        delete (msg.first)->fieldtext;
        delete msg.first;
		msg.first = temp;
	} // while
	
	msg.first = NULL;
	msg.number[0] = 0;

} // kill_message


